module lib_io_binary
  use lib_const, only: &
    CLEN_KEY, &
    DTYPE_INT1, &
    DTYPE_INT2, &
    DTYPE_INT4, &
    DTYPE_INT8, &
    DTYPE_REAL, &
    DTYPE_DBLE, &
    REC_UNDEF, &
    ENDIAN_LITTLE, &
    INT_ENDIAN_LITTLE
  use lib_log, only: &
    CODE, &
    str, &
    dgt, &
    echo, &
    edbg, &
    eerr
  use lib_base, only: &
    msg_io_error, &
    msg_invalid_value, &
    msg_unexpected_condition
  use lib_io_base, only: &
    byte_dtype, &
    int_endian
  use lib_io_file, only: &
    filesize, &
    remove, &
    make_empty_file
  implicit none
  private
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  public :: read_binary
  public :: read_binary_from2d

  public :: write_binary
  !-------------------------------------------------------------
  ! Interfaces (public)
  !-------------------------------------------------------------
  interface read_binary
    module procedure rb_int1_1d
    module procedure rb_int2_1d
    module procedure rb_int4_1d
    module procedure rb_int8_1d
    module procedure rb_real_1d
    module procedure rb_dble_1d

    module procedure rb_int1_2d
    module procedure rb_int2_2d
    module procedure rb_int4_2d
    module procedure rb_int8_2d
    module procedure rb_real_2d
    module procedure rb_dble_2d
  end interface read_binary

  interface read_binary_from2d
    module procedure rb_int8_from2d
    module procedure rb_dble_from2d
  end interface read_binary_from2d

  interface write_binary
    module procedure wb_int1_1d
    module procedure wb_int2_1d
    module procedure wb_int4_1d
    module procedure wb_int8_1d
    module procedure wb_real_1d
    module procedure wb_dble_1d

    module procedure wb_int1_2d
    module procedure wb_int2_2d
    module procedure wb_int4_2d
    module procedure wb_int8_2d
    module procedure wb_real_2d
    module procedure wb_dble_2d
  end interface write_binary
  !-------------------------------------------------------------
  ! Interfaces (private)
  !-------------------------------------------------------------
  interface read_row
    module procedure read_row_for2d_int1
    module procedure read_row_for2d_int2
    module procedure read_row_for2d_int4
    module procedure read_row_for2d_int8
    module procedure read_row_for2d_real
    module procedure read_row_for2d_dble
  end interface read_row

  interface write_row
    module procedure write_row_for1d_int1
    module procedure write_row_for1d_int2
    module procedure write_row_for1d_int4
    module procedure write_row_for1d_int8
    module procedure write_row_for1d_real
    module procedure write_row_for1d_dble

    module procedure write_row_for2d_int1
    module procedure write_row_for2d_int2
    module procedure write_row_for2d_int4
    module procedure write_row_for2d_int8
    module procedure write_row_for2d_real
    module procedure write_row_for2d_dble
  end interface write_row

  interface fill_empty_space
    module procedure fill_empty_space_int1
    module procedure fill_empty_space_int2
    module procedure fill_empty_space_int4
    module procedure fill_empty_space_int8
    module procedure fill_empty_space_real
    module procedure fill_empty_space_dble
  end interface fill_empty_space
  !-------------------------------------------------------------
  ! Module variables (private)
  !-------------------------------------------------------------
  type cmn_
    integer(1), pointer :: arr_int1(:)
    integer(2), pointer :: arr_int2(:)
    integer(4), pointer :: arr_int4(:)
    integer(8), pointer :: arr_int8(:)
    real(4)   , pointer :: arr_real(:)
    real(8)   , pointer :: arr_dble(:)
    integer(8) :: imax_int1 = 0_8
    integer(8) :: imax_int2 = 0_8
    integer(8) :: imax_int4 = 0_8
    integer(8) :: imax_int8 = 0_8
    integer(8) :: imax_real = 0_8
    integer(8) :: imax_dble = 0_8
  end type
  type(cmn_) :: cmn

  logical     , parameter :: debug = .false.
  character(8), parameter :: opt_debug = '-p'
!---------------------------------------------------------------
contains
!===============================================================
!
!===============================================================
subroutine rb_int1_1d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(1)  , intent(out) :: dat(:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)            , pointer :: offset_set
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int1_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT1  ! modified
  rec_ = 1
  sz_ = size(dat,kind=8)
  lb_ = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat)

  ub_ = lb_ + size(dat) - 1_8

  allocate(offset_set)
  offset_set = byte_dtype(dtype_) * (sz_*(rec_-1) + lb_-1_8)

  allocate(offset_cur_tail)
  offset_cur_tail = byte_dtype(dtype_) * (sz_ - ub_)

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Read ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set     : '//str(offset_set)//&
            '\n       cur_tail: '//str(offset_cur_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Move to the head position to output.
  !-------------------------------------------------------------
  call c_fseek_set(offset_set, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('set', f_, offset_set, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call c_fread_int1(dat_int1, i1max, little_, ios_)

    dat = int(dat_int1,1)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call c_fread_int2(dat_int2, i1max, little_, ios_)

    dat = int(dat_int2,1)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call c_fread_int4(dat_int4, i1max, little_, ios_)

    dat = int(dat_int4,1)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call c_fread_int8(dat_int8, i1max, little_, ios_)

    dat = int(dat_int8,1)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call c_fread_real(dat_real, i1max, little_, ios_)

    dat = int(dat_real,1)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call c_fread_dble(dat_dble, i1max, little_, ios_)

    dat = int(dat_dble,1)  ! modified
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call raise_error_fread(dtype, f_, i1max, little_, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  call c_fseek_cur(offset_cur_tail, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('cur', f_, offset_cur_tail, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%ENT, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int1_1d
!===============================================================
!
!===============================================================
subroutine rb_int2_1d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(2)  , intent(out) :: dat(:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)            , pointer :: offset_set
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int2_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT2  ! modified
  rec_ = 1
  sz_ = size(dat,kind=8)
  lb_ = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat)

  ub_ = lb_ + size(dat) - 1_8

  allocate(offset_set)
  offset_set = byte_dtype(dtype_) * (sz_*(rec_-1) + lb_-1_8)

  allocate(offset_cur_tail)
  offset_cur_tail = byte_dtype(dtype_) * (sz_ - ub_)

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Read ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set     : '//str(offset_set)//&
            '\n       cur_tail: '//str(offset_cur_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Move to the head position to output.
  !-------------------------------------------------------------
  call c_fseek_set(offset_set, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('set', f_, offset_set, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call c_fread_int1(dat_int1, i1max, little_, ios_)

    dat = int(dat_int1,2)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call c_fread_int2(dat_int2, i1max, little_, ios_)

    dat = int(dat_int2,2)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call c_fread_int4(dat_int4, i1max, little_, ios_)

    dat = int(dat_int4,2)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call c_fread_int8(dat_int8, i1max, little_, ios_)

    dat = int(dat_int8,2)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call c_fread_real(dat_real, i1max, little_, ios_)

    dat = int(dat_real,2)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call c_fread_dble(dat_dble, i1max, little_, ios_)

    dat = int(dat_dble,2)  ! modified
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call raise_error_fread(dtype, f_, i1max, little_, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  call c_fseek_cur(offset_cur_tail, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('cur', f_, offset_cur_tail, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%ENT, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int2_1d
!===============================================================
!
!===============================================================
subroutine rb_int4_1d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(4)  , intent(out) :: dat(:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)            , pointer :: offset_set
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int4_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT4  ! modified
  rec_ = 1
  sz_ = size(dat,kind=8)
  lb_ = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat)

  ub_ = lb_ + size(dat) - 1_8

  allocate(offset_set)
  offset_set = byte_dtype(dtype_) * (sz_*(rec_-1) + lb_-1_8)

  allocate(offset_cur_tail)
  offset_cur_tail = byte_dtype(dtype_) * (sz_ - ub_)

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'


  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Read ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set     : '//str(offset_set)//&
            '\n       cur_tail: '//str(offset_cur_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Move to the head position to output.
  !-------------------------------------------------------------
  call c_fseek_set(offset_set, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('set', f_, offset_set, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call c_fread_int1(dat_int1, i1max, little_, ios_)

    dat = int(dat_int1,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call c_fread_int2(dat_int2, i1max, little_, ios_)

    dat = int(dat_int2,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call c_fread_int4(dat_int4, i1max, little_, ios_)

    dat = int(dat_int4,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call c_fread_int8(dat_int8, i1max, little_, ios_)

    dat = int(dat_int8,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call c_fread_real(dat_real, i1max, little_, ios_)

    dat = int(dat_real,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call c_fread_dble(dat_dble, i1max, little_, ios_)

    dat = int(dat_dble,4)  ! modified
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call raise_error_fread(dtype, f_, i1max, little_, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  call c_fseek_cur(offset_cur_tail, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('cur', f_, offset_cur_tail, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%ENT, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int4_1d
!===============================================================
!
!===============================================================
subroutine rb_int8_1d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(8)  , intent(out) :: dat(:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)            , pointer :: offset_set
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int8_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT8  ! modified
  rec_ = 1
  sz_ = size(dat,kind=8)
  lb_ = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat)

  ub_ = lb_ + size(dat) - 1_8

  allocate(offset_set)
  offset_set = byte_dtype(dtype_) * (sz_*(rec_-1) + lb_-1_8)

  allocate(offset_cur_tail)
  offset_cur_tail = byte_dtype(dtype_) * (sz_ - ub_)

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Read ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set     : '//str(offset_set)//&
            '\n       cur_tail: '//str(offset_cur_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Move to the head position to output.
  !-------------------------------------------------------------
  call c_fseek_set(offset_set, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('set', f_, offset_set, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call c_fread_int1(dat_int1, i1max, little_, ios_)

    dat = int(dat_int1,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call c_fread_int2(dat_int2, i1max, little_, ios_)

    dat = int(dat_int2,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call c_fread_int4(dat_int4, i1max, little_, ios_)

    dat = int(dat_int4,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call c_fread_int8(dat_int8, i1max, little_, ios_)

    dat = int(dat_int8,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call c_fread_real(dat_real, i1max, little_, ios_)

    dat = int(dat_real,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call c_fread_dble(dat_dble, i1max, little_, ios_)

    dat = int(dat_dble,8)  ! modified
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call raise_error_fread(dtype, f_, i1max, little_, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  call c_fseek_cur(offset_cur_tail, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('cur', f_, offset_cur_tail, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%ENT, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int8_1d
!===============================================================
!
!===============================================================
subroutine rb_real_1d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  real(4)     , intent(out) :: dat(:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)            , pointer :: offset_set
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_real_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_REAL  ! modified
  rec_ = 1
  sz_ = size(dat,kind=8)
  lb_ = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat)

  ub_ = lb_ + size(dat) - 1_8

  allocate(offset_set)
  offset_set = byte_dtype(dtype_) * (sz_*(rec_-1) + lb_-1_8)

  allocate(offset_cur_tail)
  offset_cur_tail = byte_dtype(dtype_) * (sz_ - ub_)

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Read ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set     : '//str(offset_set)//&
            '\n       cur_tail: '//str(offset_cur_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Move to the head position to output.
  !-------------------------------------------------------------
  call c_fseek_set(offset_set, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('set', f_, offset_set, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call c_fread_int1(dat_int1, i1max, little_, ios_)

    dat = real(dat_int1,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call c_fread_int2(dat_int2, i1max, little_, ios_)

    dat = real(dat_int2,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call c_fread_int4(dat_int4, i1max, little_, ios_)

    dat = real(dat_int4,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call c_fread_int8(dat_int8, i1max, little_, ios_)

    dat = real(dat_int8,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call c_fread_real(dat_real, i1max, little_, ios_)

    dat = real(dat_real,4)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call c_fread_dble(dat_dble, i1max, little_, ios_)

    dat = real(dat_dble,4)  ! modified
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call raise_error_fread(dtype, f_, i1max, little_, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  call c_fseek_cur(offset_cur_tail, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('cur', f_, offset_cur_tail, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%ENT, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_real_1d
!===============================================================
!
!===============================================================
subroutine rb_dble_1d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  real(8)     , intent(out) :: dat(:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)            , pointer :: offset_set
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_dble_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_DBLE  ! modified
  rec_ = 1
  sz_ = size(dat,kind=8)
  lb_ = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat)

  ub_ = lb_ + size(dat) - 1_8

  allocate(offset_set)
  offset_set = byte_dtype(dtype_) * (sz_*(rec_-1) + lb_-1_8)

  allocate(offset_cur_tail)
  offset_cur_tail = byte_dtype(dtype_) * (sz_ - ub_)

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Read ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set     : '//str(offset_set)//&
            '\n       cur_tail: '//str(offset_cur_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Move to the head position to output.
  !-------------------------------------------------------------
  call c_fseek_set(offset_set, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('set', f_, offset_set, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call c_fread_int1(dat_int1, i1max, little_, ios_)

    dat = real(dat_int1,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call c_fread_int2(dat_int2, i1max, little_, ios_)

    dat = real(dat_int2,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call c_fread_int4(dat_int4, i1max, little_, ios_)

    dat = real(dat_int4,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call c_fread_int8(dat_int8, i1max, little_, ios_)

    dat = real(dat_int8,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call c_fread_real(dat_real, i1max, little_, ios_)

    dat = real(dat_real,8)  ! modified
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call c_fread_dble(dat_dble, i1max, little_, ios_)

    dat = real(dat_dble,8)  ! modified
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call raise_error_fread(dtype, f_, i1max, little_, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  call c_fseek_cur(offset_cur_tail, ios_)
  if( ios_ /= 0 )then
    call raise_error_fseek('cur', f_, offset_cur_tail, opt)
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%ENT, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_dble_1d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine rb_int1_2d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(1)  , intent(out) :: dat(:,:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8) :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int1_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT1  ! modified
  rec_ = 1
  sz_(:) = shape(dat)
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = byte_dtype(dtype_) * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = byte_dtype(dtype_) * (sz_(1) - size(dat,1))

  allocate(offset_cur_tail)
  offset_cur_tail = product(sz_) - (sz_(1)*(ub_(2)-1_8) + ub_(1))

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Read ('//str((/lb_(1),ub_(1)/),':')//&
                  ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set: '//str(offset_set)//&
            '\n       cur: '//str(offset_cur))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, i2max
      call read_row(dat_int1, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int1(:), 1)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, i2max
      call read_row(dat_int2, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int2(:), 1)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, i2max
      call read_row(dat_int4, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int4(:), 1)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, i2max
      call read_row(dat_int8, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int8(:), 1)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, i2max
      call read_row(dat_real, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_real(:), 1)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, i2max
      call read_row(dat_dble, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_dble(:), 1)  ! modified
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int1_2d
!===============================================================
!
!===============================================================
subroutine rb_int2_2d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(2)  , intent(out) :: dat(:,:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8) :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int2_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT2  ! modified
  rec_ = 1
  sz_(:) = shape(dat)
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = byte_dtype(dtype_) * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = byte_dtype(dtype_) * (sz_(1) - size(dat,1))

  allocate(offset_cur_tail)
  offset_cur_tail = product(sz_) - (sz_(1)*(ub_(2)-1_8) + ub_(1))

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Read ('//str((/lb_(1),ub_(1)/),':')//&
                  ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set: '//str(offset_set)//&
            '\n       cur: '//str(offset_cur))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, i2max
      call read_row(dat_int1, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int1(:), 2)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, i2max
      call read_row(dat_int2, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int2(:), 2)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, i2max
      call read_row(dat_int4, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int4(:), 2)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, i2max
      call read_row(dat_int8, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int8(:), 2)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, i2max
      call read_row(dat_real, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_real(:), 2)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, i2max
      call read_row(dat_dble, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_dble(:), 2)  ! modified
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int2_2d
!===============================================================
!
!===============================================================
subroutine rb_int4_2d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(4)  , intent(out) :: dat(:,:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8) :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int4_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT4  ! modified
  rec_ = 1
  sz_(:) = shape(dat)
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = byte_dtype(dtype_) * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = byte_dtype(dtype_) * (sz_(1) - size(dat,1))

  allocate(offset_cur_tail)
  offset_cur_tail = product(sz_) - (sz_(1)*(ub_(2)-1_8) + ub_(1))

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Read ('//str((/lb_(1),ub_(1)/),':')//&
                  ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set: '//str(offset_set)//&
            '\n       cur: '//str(offset_cur))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, i2max
      call read_row(dat_int1, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int1(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, i2max
      call read_row(dat_int2, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int2(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, i2max
      call read_row(dat_int4, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int4(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, i2max
      call read_row(dat_int8, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int8(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, i2max
      call read_row(dat_real, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_real(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, i2max
      call read_row(dat_dble, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_dble(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int4_2d
!===============================================================
!
!===============================================================
subroutine rb_int8_2d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(8)  , intent(out) :: dat(:,:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8) :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_int8_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT8  ! modified
  rec_ = 1
  sz_(:) = shape(dat)
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = byte_dtype(dtype_) * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = byte_dtype(dtype_) * (sz_(1) - size(dat,1))

  allocate(offset_cur_tail)
  offset_cur_tail = product(sz_) - (sz_(1)*(ub_(2)-1_8) + ub_(1))

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Read ('//str((/lb_(1),ub_(1)/),':')//&
                  ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set: '//str(offset_set)//&
            '\n       cur: '//str(offset_cur))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, i2max
      call read_row(dat_int1, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int1(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, i2max
      call read_row(dat_int2, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int2(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, i2max
      call read_row(dat_int4, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int4(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, i2max
      call read_row(dat_int8, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_int8(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, i2max
      call read_row(dat_real, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_real(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, i2max
      call read_row(dat_dble, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = int(dat_dble(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_int8_2d
!===============================================================
!
!===============================================================
subroutine rb_real_2d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  real(4)     , intent(out) :: dat(:,:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8) :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_real_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_REAL  ! modified
  rec_ = 1
  sz_(:) = shape(dat)
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = byte_dtype(dtype_) * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = byte_dtype(dtype_) * (sz_(1) - size(dat,1))

  allocate(offset_cur_tail)
  offset_cur_tail = product(sz_) - (sz_(1)*(ub_(2)-1_8) + ub_(1))

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Read ('//str((/lb_(1),ub_(1)/),':')//&
                  ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set: '//str(offset_set)//&
            '\n       cur: '//str(offset_cur))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, i2max
      call read_row(dat_int1, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int1(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, i2max
      call read_row(dat_int2, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int2(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, i2max
      call read_row(dat_int4, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int4(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, i2max
      call read_row(dat_int8, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int8(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, i2max
      call read_row(dat_real, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_real(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, i2max
      call read_row(dat_dble, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_dble(:), 4)  ! modified
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call finalize(.true.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_real_2d
!===============================================================
!
!===============================================================
subroutine rb_dble_2d(dat, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  real(8)     , intent(out) :: dat(:,:)  ! modified
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)            , pointer :: offset_cur_tail
  integer(4)            , pointer :: ios_
  character(8) :: opt

  integer(1), pointer :: dat_int1(:)
  integer(2), pointer :: dat_int2(:)
  integer(4), pointer :: dat_int4(:)
  integer(8), pointer :: dat_int8(:)
  real(4)   , pointer :: dat_real(:)
  real(8)   , pointer :: dat_dble(:)

  call echo(CODE%BGN, 'read_binary__MP__rb_dble_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_DBLE  ! modified
  rec_ = 1
  sz_(:) = shape(dat)
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(sz) ) sz_ = sz
  if( present(lb) ) lb_ = lb

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = byte_dtype(dtype_) * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = byte_dtype(dtype_) * (sz_(1) - size(dat,1))

  allocate(offset_cur_tail)
  offset_cur_tail = product(sz_) - (sz_(1)*(ub_(2)-1_8) + ub_(1))

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Read ('//str((/lb_(1),ub_(1)/),':')//&
                  ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set: '//str(offset_set)//&
            '\n       cur: '//str(offset_cur))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb', 2, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Allocate a pointer for reading.
  !-------------------------------------------------------------
  call alloc_ptr_1d(dtype_, i1max, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  ! Read data.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, i2max
      call read_row(dat_int1, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int1(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, i2max
      call read_row(dat_int2, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int2(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, i2max
      call read_row(dat_int4, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int4(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, i2max
      call read_row(dat_int8, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_int8(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, i2max
      call read_row(dat_real, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_real(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, i2max
      call read_row(dat_dble, i1max, i2max, i2, &
                    offset_set, offset_cur, offset_cur_tail, &
                    f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit

      dat(:,i2) = real(dat_dble(:), 8)  ! modified
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  if( ios_ /= 0 )then
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.false.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)

  call dealloc_ptr_1d(dtype_, dat_int1, dat_int2, dat_int4, dat_int8, dat_real, dat_dble)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine rb_dble_2d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine rb_int8_from2d(dat, nx, ny, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  integer(8)  , intent(out) :: dat(:)
  integer(8)  , intent(in)  :: nx, ny
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(clen_key) :: dtype_
  integer :: rec_
  character(clen_key) :: endian_
  integer(8) :: sz_(2)
  integer(8) :: lb_(2)
  integer :: ios_

  integer(8), allocatable :: dat_2d(:,:)
  character(8) :: opt

  call echo(code%bgn, 'read_binary_from2d__MP__rb_int8_from_2d', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dtype_ = DTYPE_INT8
  rec_ = 1
  endian_ = ENDIAN_LITTLE
  sz_(1) = nx
  sz_(2) = ny
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(endian) ) endian_ = endian
  if( present(sz) ) sz_(:) = sz(:)
  if( present(lb) ) lb_(:) = lb(:)

  if( present(ios) ) opt = '-q -c'

  if( size(dat) /= nx*ny )then
    call eerr(str(msg_unexpected_condition())//&
            '\n  size(dat) /= nx*ny'//&
            '\n  size(dat): '//str(size(dat))//&
            '\n  nx*ny: '//str(nx*ny)//' ((nx,ny): ('//str((/nx,ny/),', ')//')')
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(dat_2d(nx,ny))

  call rb_int8_2d(dat_2d, f, dtype_, rec_, endian_, sz_, lb_, ios_)

  if( ios_ == 0 )then
    dat(:) = reshape(dat_2d,(/size(dat)/))
  else
    call eerr(str(msg_io_error()), opt)
  endif

  deallocate(dat_2d)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(ios) ) ios = ios_
  !-------------------------------------------------------------
  call echo(code%ret)
end subroutine rb_int8_from2d
!===============================================================
!
!===============================================================
subroutine rb_dble_from2d(dat, nx, ny, f, dtype, rec, endian, sz, lb, ios)
  implicit none
  real(8)     , intent(out) :: dat(:)
  integer(8)  , intent(in)  :: nx, ny
  character(*), intent(in)  :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  integer     , intent(out), optional :: ios

  character(clen_key) :: dtype_
  integer :: rec_
  character(clen_key) :: endian_
  integer(8) :: sz_(2)
  integer(8) :: lb_(2)
  integer :: ios_

  real(8), allocatable :: dat_2d(:,:)
  character(8) :: opt

  call echo(code%bgn, 'read_binary_from2d__MP__rb_dble_from_2d', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dtype_ = DTYPE_DBLE
  rec_ = 1
  endian_ = ENDIAN_LITTLE
  sz_(1) = nx
  sz_(2) = ny
  lb_(:) = 1_8

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(endian) ) endian_ = endian
  if( present(sz) ) sz_(:) = sz(:)
  if( present(lb) ) lb_(:) = lb(:)

  if( present(ios) ) opt = '-q -c'

  if( size(dat) /= nx*ny )then
    call eerr(str(msg_unexpected_condition())//&
            '\n  size(dat) /= nx*ny'//&
            '\n  size(dat): '//str(size(dat))//&
            '\n  nx*ny: '//str(nx*ny)//' ((nx,ny): ('//str((/nx,ny/),', ')//')')
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(dat_2d(nx,ny))

  call rb_dble_2d(dat_2d, f, dtype_, rec_, endian_, sz_, lb_, ios_)

  if( ios_ == 0 )then
    dat(:) = reshape(dat_2d,(/size(dat)/))
  else
    call eerr(str(msg_io_error()), opt)
  endif

  deallocate(dat_2d)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(ios) ) ios = ios_
  !-------------------------------------------------------------
  call echo(code%ret)
end subroutine rb_dble_from2d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine read_row_for2d_int1(&
    dat, i1max, i2max, i2, &
    offset_set, offset_cur, offset_cur_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(1)  , pointer    :: dat(:)  ! modified
  integer(8)  , pointer    :: i1max
  integer(8)  , intent(in) :: i2max, i2
  integer(8)  , pointer    :: offset_set, offset_cur, offset_cur_tail
  character(*), intent(in) :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'read_row__MP__read_row_for2d_int1', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == 1_8 )then
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_cur(offset_cur, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  call c_fread_int1(dat, i1max, little, ios)  ! modified
  if( ios /= 0 )then
    call raise_error_fread(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  if( i2 == i2max )then
    call c_fseek_cur(offset_cur_tail, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur_tail, opt)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine read_row_for2d_int1
!===============================================================
!
!===============================================================
subroutine read_row_for2d_int2(&
    dat, i1max, i2max, i2, &
    offset_set, offset_cur, offset_cur_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(2)  , pointer    :: dat(:)  ! modified
  integer(8)  , pointer    :: i1max
  integer(8)  , intent(in) :: i2max, i2
  integer(8)  , pointer    :: offset_set, offset_cur, offset_cur_tail
  character(*), intent(in) :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'read_row__MP__read_row_for2d_int2', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == 1_8 )then
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_cur(offset_cur, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  call c_fread_int2(dat, i1max, little, ios)  ! modified
  if( ios /= 0 )then
    call raise_error_fread(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  if( i2 == i2max )then
    call c_fseek_cur(offset_cur_tail, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur_tail, opt)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine read_row_for2d_int2
!===============================================================
!
!===============================================================
subroutine read_row_for2d_int4(&
    dat, i1max, i2max, i2, &
    offset_set, offset_cur, offset_cur_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(4)  , pointer    :: dat(:)  ! modified
  integer(8)  , pointer    :: i1max
  integer(8)  , intent(in) :: i2max, i2
  integer(8)  , pointer    :: offset_set, offset_cur, offset_cur_tail
  character(*), intent(in) :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'read_row__MP__read_row_for2d_int4', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == 1_8 )then
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_cur(offset_cur, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  call c_fread_int4(dat, i1max, little, ios)  ! modified
  if( ios /= 0 )then
    call raise_error_fread(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  if( i2 == i2max )then
    call c_fseek_cur(offset_cur_tail, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur_tail, opt)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine read_row_for2d_int4
!===============================================================
!
!===============================================================
subroutine read_row_for2d_int8(&
    dat, i1max, i2max, i2, &
    offset_set, offset_cur, offset_cur_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(8)  , pointer    :: dat(:)  ! modified
  integer(8)  , pointer    :: i1max
  integer(8)  , intent(in) :: i2max, i2
  integer(8)  , pointer    :: offset_set, offset_cur, offset_cur_tail
  character(*), intent(in) :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'read_row__MP__read_row_for2d_int8', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == 1_8 )then
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_cur(offset_cur, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  call c_fread_int8(dat, i1max, little, ios)  ! modified
  if( ios /= 0 )then
    call raise_error_fread(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  if( i2 == i2max )then
    call c_fseek_cur(offset_cur_tail, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur_tail, opt)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine read_row_for2d_int8
!===============================================================
!
!===============================================================
subroutine read_row_for2d_real(&
    dat, i1max, i2max, i2, &
    offset_set, offset_cur, offset_cur_tail, &
    f, dtype, little, ios, opt)
  implicit none
  real(4)     , pointer    :: dat(:)  ! modified
  integer(8)  , pointer    :: i1max
  integer(8)  , intent(in) :: i2max, i2
  integer(8)  , pointer    :: offset_set, offset_cur, offset_cur_tail
  character(*), intent(in) :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'read_row__MP__read_row_for2d_real', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == 1_8 )then
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_cur(offset_cur, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  call c_fread_real(dat, i1max, little, ios)  ! modified
  if( ios /= 0 )then
    call raise_error_fread(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  if( i2 == i2max )then
    call c_fseek_cur(offset_cur_tail, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur_tail, opt)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine read_row_for2d_real
!===============================================================
!
!===============================================================
subroutine read_row_for2d_dble(&
    dat, i1max, i2max, i2, &
    offset_set, offset_cur, offset_cur_tail, &
    f, dtype, little, ios, opt)
  implicit none
  real(8)     , pointer    :: dat(:)  ! modified
  integer(8)  , pointer    :: i1max
  integer(8)  , intent(in) :: i2max, i2
  integer(8)  , pointer    :: offset_set, offset_cur, offset_cur_tail
  character(*), intent(in) :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'read_row__MP__read_row_for2d_dble', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == 1_8 )then
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_cur(offset_cur, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  call c_fread_dble(dat, i1max, little, ios)  ! modified
  if( ios /= 0 )then
    call raise_error_fread(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  if( i2 == i2max )then
    call c_fseek_cur(offset_cur_tail, ios)
    if( ios /= 0 )then
      call raise_error_fseek('cur', f, offset_cur_tail, opt)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine read_row_for2d_dble
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine wb_int1_1d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(1)  , intent(in) :: dat(:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  logical     , intent(in) , optional :: replace
  integer(1)  , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(1)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set
  integer(8)                      :: offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int1_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT1  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0_1  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,kind=8)

  lb_ = 1_8
  if( present(lb) ) lb_ = lb

  ub_ = lb_ + size(dat) - 1_8

  sz_ = ub_
  if( present(sz) ) sz_ = sz

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  offset_set = wbyte * (sz_*(rec_-1) + lb_-1_8)

  offset_tail = wbyte * sz_ * rec_

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Write ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call write_row(int(dat(:),1), int(fill_,1), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call write_row(int(dat(:),2), int(fill_,2), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call write_row(int(dat(:),4), int(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call write_row(int(dat(:),8), int(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call write_row(real(dat(:),4), real(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call write_row(real(dat(:),8), real(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !---------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int1_1d
!===============================================================
!
!===============================================================
subroutine wb_int2_1d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(2)  , intent(in) :: dat(:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  logical     , intent(in) , optional :: replace
  integer(2)  , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(2)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set
  integer(8)                      :: offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int2_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT2
  rec_ = 1
  replace_ = .false.
  fill_ = 0_2  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,kind=8)

  lb_ = 1_8
  if( present(lb) ) lb_ = lb

  ub_ = lb_ + size(dat) - 1_8

  sz_ = ub_
  if( present(sz) ) sz_ = sz

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  offset_set = wbyte * (sz_*(rec_-1) + lb_-1_8)

  offset_tail = wbyte * sz_ * rec_

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Write ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call write_row(int(dat(:),1), int(fill_,1), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call write_row(int(dat(:),2), int(fill_,2), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call write_row(int(dat(:),4), int(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call write_row(int(dat(:),8), int(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call write_row(real(dat(:),4), real(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call write_row(real(dat(:),8), real(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !---------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int2_1d
!===============================================================
!
!===============================================================
subroutine wb_int4_1d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(4)  , intent(in) :: dat(:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  logical     , intent(in) , optional :: replace
  integer(4)  , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(4)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set
  integer(8)                      :: offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int4_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT4  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0_4  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,kind=8)

  lb_ = 1_8
  if( present(lb) ) lb_ = lb

  ub_ = lb_ + size(dat) - 1_8

  sz_ = ub_
  if( present(sz) ) sz_ = sz

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  offset_set = wbyte * (sz_*(rec_-1) + lb_-1_8)

  offset_tail = wbyte * sz_ * rec_

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Write ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call write_row(int(dat(:),1), int(fill_,1), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call write_row(int(dat(:),2), int(fill_,2), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call write_row(int(dat(:),4), int(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call write_row(int(dat(:),8), int(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call write_row(real(dat(:),4), real(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call write_row(real(dat(:),8), real(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !---------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int4_1d
!===============================================================
!
!===============================================================
subroutine wb_int8_1d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(8)  , intent(in) :: dat(:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  logical     , intent(in) , optional :: replace
  integer(8)  , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(8)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set
  integer(8)                      :: offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int8_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT8  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0_8  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,kind=8)

  lb_ = 1_8
  if( present(lb) ) lb_ = lb

  ub_ = lb_ + size(dat) - 1_8

  sz_ = ub_
  if( present(sz) ) sz_ = sz

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  offset_set = wbyte * (sz_*(rec_-1) + lb_-1_8)

  offset_tail = wbyte * sz_ * rec_

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Write ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call write_row(int(dat(:),1), int(fill_,1), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call write_row(int(dat(:),2), int(fill_,2), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call write_row(int(dat(:),4), int(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call write_row(int(dat(:),8), int(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call write_row(real(dat(:),4), real(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call write_row(real(dat(:),8), real(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !---------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int8_1d
!===============================================================
!
!===============================================================
subroutine wb_real_1d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  real(4)     , intent(in) :: dat(:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  logical     , intent(in) , optional :: replace
  real(4)     , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  logical                         :: replace_
  real(4)                         :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set
  integer(8)                      :: offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_real_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_REAL
  rec_ = 1
  replace_ = .false.
  fill_ = 0.e0  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,kind=8)

  lb_ = 1_8
  if( present(lb) ) lb_ = lb

  ub_ = lb_ + size(dat) - 1_8

  sz_ = ub_
  if( present(sz) ) sz_ = sz

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  offset_set = wbyte * (sz_*(rec_-1) + lb_-1_8)

  offset_tail = wbyte * sz_ * rec_

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Write ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call write_row(int(dat(:),1), int(fill_,1), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call write_row(int(dat(:),2), int(fill_,2), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call write_row(int(dat(:),4), int(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call write_row(int(dat(:),8), int(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call write_row(real(dat(:),4), real(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call write_row(real(dat(:),8), real(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !---------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_real_1d
!===============================================================
!
!===============================================================
subroutine wb_dble_1d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  real(8)     , intent(in) :: dat(:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz
  integer(8)  , intent(in) , optional :: lb
  logical     , intent(in) , optional :: replace
  real(8)     , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_
  integer(8)                      :: lb_, ub_
  integer(4)            , pointer :: little_
  logical                         :: replace_
  real(8)                         :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set
  integer(8)                      :: offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_dble_1d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_DBLE
  rec_ = 1
  replace_ = .false.
  fill_ = 0.d0  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,kind=8)

  lb_ = 1_8
  if( present(lb) ) lb_ = lb

  ub_ = lb_ + size(dat) - 1_8

  sz_ = ub_
  if( present(sz) ) sz_ = sz

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  offset_set = wbyte * (sz_*(rec_-1) + lb_-1_8)

  offset_tail = wbyte * sz_ * rec_

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: '//str(sz_))
    call edbg('Write ('//str((/lb_,ub_/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), (/sz_/), (/lb_/), (/ub_/), ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    call write_row(int(dat(:),1), int(fill_,1), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    call write_row(int(dat(:),2), int(fill_,2), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    call write_row(int(dat(:),4), int(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    call write_row(int(dat(:),8), int(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    call write_row(real(dat(:),4), real(fill_,4), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    call write_row(real(dat(:),8), real(fill_,8), wbyte, i1max, &
                   fsize, offset_set, offset_tail, &
                   f_, dtype_, little_, ios_, opt)
  !---------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_dble_1d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine wb_int1_2d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(1)  , intent(in) :: dat(:,:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  logical     , intent(in) , optional :: replace
  integer(1)  , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(1)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)                      :: offset_now, offset_next, offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int1_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT1
  rec_ = 1
  replace_ = .false.
  fill_ = 0_1  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  lb_(:) = 1_8
  if( present(lb) ) lb_(:) = lb(:)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  sz_(:) = ub_(:)
  if( present(sz) ) sz_(:) = sz(:)

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = wbyte * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = wbyte * (sz_(1) - size(dat,1))

  offset_tail = product(sz_) * rec_ * wbyte

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Write ('//str((/lb_(1),ub_(1)/),':')//&
                   ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       cur : '//str(offset_cur)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  offset_now = 0_8
  offset_next = offset_set

  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),1), int(fill_,1), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),2), int(fill_,2), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),4), int(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),8), int(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),4), real(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),8), real(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int1_2d
!===============================================================
!
!===============================================================
subroutine wb_int2_2d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(2)  , intent(in) :: dat(:,:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  logical     , intent(in) , optional :: replace
  integer(2)  , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(2)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)                      :: offset_now, offset_next, offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int2_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT2  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0_2  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  lb_(:) = 1_8
  if( present(lb) ) lb_(:) = lb(:)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  sz_(:) = ub_(:)
  if( present(sz) ) sz_(:) = sz(:)

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = wbyte * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = wbyte * (sz_(1) - size(dat,1))

  offset_tail = product(sz_) * rec_ * wbyte

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Write ('//str((/lb_(1),ub_(1)/),':')//&
                   ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       cur : '//str(offset_cur)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  offset_now = 0_8
  offset_next = offset_set

  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),1), int(fill_,1), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),2), int(fill_,2), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),4), int(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),8), int(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),4), real(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),8), real(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int2_2d
!===============================================================
!
!===============================================================
subroutine wb_int4_2d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(4)  , intent(in) :: dat(:,:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  logical     , intent(in) , optional :: replace
  integer(4)  , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(4)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)                      :: offset_now, offset_next, offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int4_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT4  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0_4  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  lb_(:) = 1_8
  if( present(lb) ) lb_(:) = lb(:)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  sz_(:) = ub_(:)
  if( present(sz) ) sz_(:) = sz(:)

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = wbyte * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = wbyte * (sz_(1) - size(dat,1))

  offset_tail = product(sz_) * rec_ * wbyte

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Write ('//str((/lb_(1),ub_(1)/),':')//&
                   ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       cur : '//str(offset_cur)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  offset_now = 0_8
  offset_next = offset_set

  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),1), int(fill_,1), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),2), int(fill_,2), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),4), int(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),8), int(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),4), real(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),8), real(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int4_2d
!===============================================================
!
!===============================================================
subroutine wb_int8_2d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  integer(8)   , intent(in) :: dat(:,:)  ! modified
  character(*) , intent(in) :: f
  character(*) , intent(in) , optional :: dtype
  integer      , intent(in) , optional :: rec
  character(*) , intent(in) , optional :: endian
  integer(8)   , intent(in) , optional :: sz(:)
  integer(8)   , intent(in) , optional :: lb(:)
  logical      , intent(in) , optional :: replace
  integer(8)   , intent(in) , optional :: fill  ! modified
  integer      , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  logical                         :: replace_
  integer(8)                      :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)                      :: offset_now, offset_next, offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_int8_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_INT8  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0_8  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  lb_(:) = 1_8
  if( present(lb) ) lb_(:) = lb(:)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  sz_(:) = ub_(:)
  if( present(sz) ) sz_(:) = sz(:)

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = wbyte * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = wbyte * (sz_(1) - size(dat,1))

  offset_tail = product(sz_) * rec_ * wbyte

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Write ('//str((/lb_(1),ub_(1)/),':')//&
                   ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       cur : '//str(offset_cur)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  offset_now = 0_8
  offset_next = offset_set

  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),1), int(fill_,1), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),2), int(fill_,2), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),4), int(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),8), int(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),4), real(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),8), real(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_int8_2d
!===============================================================
!
!===============================================================
subroutine wb_real_2d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  real(4)     , intent(in) :: dat(:,:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  logical     , intent(in) , optional :: replace
  real(4)     , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  logical                         :: replace_
  real(4)                         :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)                      :: offset_now, offset_next, offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_real_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_REAL  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0.e0  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  lb_(:) = 1_8
  if( present(lb) ) lb_(:) = lb(:)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  sz_(:) = ub_(:)
  if( present(sz) ) sz_(:) = sz(:)

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = wbyte * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = wbyte * (sz_(1) - size(dat,1))

  offset_tail = product(sz_) * rec_ * wbyte

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Write ('//str((/lb_(1),ub_(1)/),':')//&
                   ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       cur : '//str(offset_cur)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  offset_now = 0_8
  offset_next = offset_set

  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),1), int(fill_,1), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),2), int(fill_,2), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),4), int(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),8), int(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),4), real(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),8), real(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_real_2d
!===============================================================
!
!===============================================================
subroutine wb_dble_2d(dat, f, dtype, rec, endian, sz, lb, replace, fill, ios)
  implicit none
  real(8)     , intent(in) :: dat(:,:)  ! modified
  character(*), intent(in) :: f
  character(*), intent(in) , optional :: dtype
  integer     , intent(in) , optional :: rec
  character(*), intent(in) , optional :: endian
  integer(8)  , intent(in) , optional :: sz(:)
  integer(8)  , intent(in) , optional :: lb(:)
  logical     , intent(in) , optional :: replace
  real(8)     , intent(in) , optional :: fill  ! modified
  integer     , intent(out), optional :: ios

  character(len_trim(f)), pointer :: f_
  character(CLEN_KEY)             :: dtype_
  integer                         :: rec_
  integer(8)                      :: sz_(2)
  integer(8)                      :: lb_(2), ub_(2)
  integer(4)            , pointer :: little_
  logical                         :: replace_
  real(8)                         :: fill_  ! modified
  integer(8)            , pointer :: i1max
  integer(8)                      :: i2max, i2
  integer(8)                      :: wbyte
  integer(8)            , pointer :: offset_set, offset_cur
  integer(8)                      :: offset_now, offset_next, offset_tail
  integer(8)                      :: fsize
  integer(4)            , pointer :: ios_
  character(8)                    :: opt

  integer :: access

  call echo(CODE%BGN, 'write_binary__MP__wb_dble_2d', opt_debug)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( present(sz) ) call check_array_size(sz, 2, 'sz')
  if( present(lb) ) call check_array_size(lb, 2, 'lb')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  allocate(f_)
  f_ = trim(f)

  dtype_ = DTYPE_DBLE  ! modified
  rec_ = 1
  replace_ = .false.
  fill_ = 0.d0  ! modified

  if( present(dtype) ) dtype_ = dtype
  if( present(rec) ) rec_ = rec
  if( present(replace) ) replace_ = replace
  if( present(fill) ) fill_ = fill

  allocate(little_)
  little_ = INT_ENDIAN_LITTLE
  if( present(endian) ) little_ = int_endian(endian)

  allocate(i1max)
  i1max = size(dat,1,kind=8)
  i2max = size(dat,2,kind=8)

  lb_(:) = 1_8
  if( present(lb) ) lb_(:) = lb(:)

  ub_(:) = lb_(:) + shape(dat) - 1_8

  sz_(:) = ub_(:)
  if( present(sz) ) sz_(:) = sz(:)

  wbyte = byte_dtype(dtype_)

  allocate(offset_set)
  allocate(offset_cur)
  offset_set = wbyte * (product(sz_)*(rec_-1) + sz_(1)*(lb_(2)-1_8) + lb_(1)-1_8)
  offset_cur = wbyte * (sz_(1) - size(dat,1))

  offset_tail = product(sz_) * rec_ * wbyte

  allocate(ios_)
  ios_ = 0

  opt = ''
  if( present(ios) ) opt = '-q'

  if( debug )then
    call edbg('Replace: '//str(replace_))
    call edbg('Record: '//str(rec_))
    call edbg('Size of record: ('//str(sz_,', ')//')')
    call edbg('Write ('//str((/lb_(1),ub_(1)/),':')//&
                   ', '//str((/lb_(2),ub_(2)/),':')//')')
    call edbg('Offset set : '//str(offset_set)//&
            '\n       cur : '//str(offset_cur)//&
            '\n       tail: '//str(offset_tail))
  endif
  !-------------------------------------------------------------
  ! Check record number, dimension and boundaries.
  !-------------------------------------------------------------
  call check_values(rec_, shape(dat), sz_, lb_, ub_, ios_, opt)
  if( ios_ /= 0 )then
    if( present(ios) ) ios = ios_
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Remove old file.
  !-------------------------------------------------------------
  if( replace_ .and. access(f,' ') == 0 )then
    call remove(f, dir=.false., output=.false.)
  endif
  !-------------------------------------------------------------
  ! Make empty file if not exist.
  !-------------------------------------------------------------
  if( access(f,' ') /= 0 )then
    call make_empty_file(f, ios_, .false., opt)
    if( ios_ /= 0 )then
      call finalize(.false.)
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  ! Get file size.
  !-------------------------------------------------------------
  fsize = filesize(f)
  !-------------------------------------------------------------
  ! Open the file.
  !-------------------------------------------------------------
  call c_fopen(f_, len_trim(f_), 'rb+', 3, ios_)
  if( ios_ /= 0 )then
    call raise_error_fopen(f_, opt)
    call finalize(.false.)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  ! Output.
  !-------------------------------------------------------------
  offset_now = 0_8
  offset_next = offset_set

  selectcase( dtype_ )
  !-------------------------------------------------------------
  case( DTYPE_INT1 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),1), int(fill_,1), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT2 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),2), int(fill_,2), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT4 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),4), int(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_INT8 )
    do i2 = 1_8, size(dat,2)
      call write_row(int(dat(:,i2),8), int(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_REAL )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),4), real(fill_,4), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case( DTYPE_DBLE )
    do i2 = 1_8, size(dat,2)
      call write_row(real(dat(:,i2),8), real(fill_,8), wbyte, i1max, i2max, i2, &
                     fsize, offset_set, offset_cur, &
                     offset_tail, offset_now, offset_next, &
                     f_, dtype_, little_, ios_, opt)
      if( ios_ /= 0 ) exit
    enddo
  !-------------------------------------------------------------
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype_))
  endselect
  !-------------------------------------------------------------
  ! Finalize.
  !-------------------------------------------------------------
  call finalize(.true.)
  !-------------------------------------------------------------
  call echo(CODE%RET)
!---------------------------------------------------------------
contains
!---------------------------------------------------------------
subroutine finalize(file_is_opened)
  implicit none
  logical, intent(in) :: file_is_opened

  call echo(CODE%BGN, 'finalize', '-p')
  !-------------------------------------------------------------
  if( file_is_opened )then
    call c_fclose(ios_)
    if( ios_ /= 0 )then
      call raise_error_fclose(f_, opt)
    endif
  endif

  if( present(ios) ) ios = ios_

  deallocate(f_)
  deallocate(little_)
  deallocate(i1max)
  deallocate(offset_set)
  deallocate(offset_cur)
  deallocate(ios_)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine finalize
!---------------------------------------------------------------
end subroutine wb_dble_2d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine write_row_for1d_int1(&
    dat, fill, byte, i1max, &
    fsize, offset_set, offset_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(1), intent(in), target :: dat(:)  ! modified
  integer(1), intent(in)         :: fill  ! modified
  integer(8), intent(in)         :: byte
  integer(8), pointer            :: i1max
  integer(8), intent(in)    :: fsize
  integer(8), pointer       :: offset_set
  integer(8), intent(in)    :: offset_tail
  character(*), pointer    :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: offset_now
  integer(1), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for1d_int1', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_set > fsize )then
    call fill_empty_space(fsize, 0_8, offset_set, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  dat_ptr => dat

  call c_fwrite_int1(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  offset_now = offset_set + i1max * byte

  if( offset_tail > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                          f, little, ios, opt)
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for1d_int1
!===============================================================
!
!===============================================================
subroutine write_row_for1d_int2(&
    dat, fill, byte, i1max, &
    fsize, offset_set, offset_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(2), intent(in), target :: dat(:)  ! modified
  integer(2), intent(in)         :: fill  ! modified
  integer(8), intent(in)         :: byte
  integer(8), pointer            :: i1max
  integer(8), intent(in)    :: fsize
  integer(8), pointer       :: offset_set
  integer(8), intent(in)    :: offset_tail
  character(*), pointer    :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: offset_now
  integer(2), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for1d_int2', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_set > fsize )then
    call fill_empty_space(fsize, 0_8, offset_set, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  dat_ptr => dat

  call c_fwrite_int2(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  offset_now = offset_set + i1max * byte

  if( offset_tail > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                          f, little, ios, opt)
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for1d_int2
!===============================================================
!
!===============================================================
subroutine write_row_for1d_int4(&
    dat, fill, byte, i1max, &
    fsize, offset_set, offset_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(4), intent(in), target :: dat(:)  ! modified
  integer(4), intent(in)         :: fill  ! modified
  integer(8), intent(in)         :: byte
  integer(8), pointer            :: i1max
  integer(8), intent(in)    :: fsize
  integer(8), pointer       :: offset_set
  integer(8), intent(in)    :: offset_tail
  character(*), pointer    :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: offset_now
  integer(4), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for1d_int4', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_set > fsize )then
    call fill_empty_space(fsize, 0_8, offset_set, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  dat_ptr => dat

  call c_fwrite_int4(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  offset_now = offset_set + i1max * byte

  if( offset_tail > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                          f, little, ios, opt)
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for1d_int4
!===============================================================
!
!===============================================================
subroutine write_row_for1d_int8(&
    dat, fill, byte, i1max, &
    fsize, offset_set, offset_tail, &
    f, dtype, little, ios, opt)
  implicit none
  integer(8), intent(in), target :: dat(:)  ! modified
  integer(8), intent(in)         :: fill  ! modified
  integer(8), intent(in)         :: byte
  integer(8), pointer            :: i1max
  integer(8), intent(in)    :: fsize
  integer(8), pointer       :: offset_set
  integer(8), intent(in)    :: offset_tail
  character(*), pointer    :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: offset_now
  integer(8), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for1d_int8', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_set > fsize )then
    call fill_empty_space(fsize, 0_8, offset_set, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  dat_ptr => dat

  call c_fwrite_int8(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  offset_now = offset_set + i1max * byte

  if( offset_tail > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                          f, little, ios, opt)
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for1d_int8
!===============================================================
!
!===============================================================
subroutine write_row_for1d_real(&
    dat, fill, byte, i1max, &
    fsize, offset_set, offset_tail, &
    f, dtype, little, ios, opt)
  implicit none
  real(4)   , intent(in), target :: dat(:)  ! modified
  real(4)   , intent(in)         :: fill  ! modified
  integer(8), intent(in)         :: byte
  integer(8), pointer            :: i1max
  integer(8), intent(in)    :: fsize
  integer(8), pointer       :: offset_set
  integer(8), intent(in)    :: offset_tail
  character(*), pointer    :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: offset_now
  real(4), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for1d_real', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_set > fsize )then
    call fill_empty_space(fsize, 0_8, offset_set, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  dat_ptr => dat

  call c_fwrite_real(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  offset_now = offset_set + i1max * byte

  if( offset_tail > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                          f, little, ios, opt)
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for1d_real
!===============================================================
!
!===============================================================
subroutine write_row_for1d_dble(&
    dat, fill, byte, i1max, &
    fsize, offset_set, offset_tail, &
    f, dtype, little, ios, opt)
  implicit none
  real(8)   , intent(in), target :: dat(:)  ! modified
  real(8)   , intent(in)         :: fill  ! modified
  integer(8), intent(in)         :: byte
  integer(8), pointer            :: i1max
  integer(8), intent(in)    :: fsize
  integer(8), pointer       :: offset_set
  integer(8), intent(in)    :: offset_tail
  character(*), pointer    :: f
  character(*), intent(in) :: dtype
  integer(4)  , pointer    :: little
  integer(4)  , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: offset_now
  real(8), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for1d_dble', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_set > fsize )then
    call fill_empty_space(fsize, 0_8, offset_set, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    call c_fseek_set(offset_set, ios)
    if( ios /= 0 )then
      call raise_error_fseek('set', f, offset_set, opt)
      call echo(CODE%RET)
      return
    endif
  endif

  dat_ptr => dat

  call c_fwrite_dble(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif

  offset_now = offset_set + i1max * byte

  if( offset_tail > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                          f, little, ios, opt)
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for1d_dble
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine write_row_for2d_int1(&
    dat, fill, byte, i1max, i2max, i2, &
    fsize, offset_set, offset_cur, &
    offset_tail, offset_now, offset_next, &
    f, dtype, little, ios, opt)
  implicit none
  integer(1)  , intent(in), target :: dat(:)  ! modified
  integer(1)  , intent(in)         :: fill  ! modified
  integer(8)  , intent(in)         :: byte
  integer(8)  , pointer            :: i1max
  integer(8)                       :: i2max, &
                                      i2
  integer(8)  , intent(in)    :: fsize
  integer(8)  , pointer       :: offset_set, &
                                 offset_cur
  integer(8)  , intent(in)    :: offset_tail
  integer(8)  , intent(inout) :: offset_now, &
                                 offset_next
  character(*), pointer     :: f
  character(*), intent(in)  :: dtype
  integer(4)  , pointer     :: little
  integer     , pointer     :: ios
  character(*), intent(in)  :: opt

  integer(1), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for2d_int1', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_next > fsize .and. offset_next > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_next, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    if( i2 == 1_8 )then
      call c_fseek_set(offset_set, ios)
      if( ios /= 0 )then
        call raise_error_fseek('set', f, offset_set, opt)
        call echo(CODE%RET)
        return
      endif
    else
      call c_fseek_cur(offset_cur, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', f, offset_cur, opt)
        call echo(CODE%RET)
        return
      endif
    endif
  endif

  offset_now = offset_next
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dat_ptr => dat

  call c_fwrite_int1(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  offset_now = offset_now + i1max * byte
  offset_next = offset_now + offset_cur
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == i2max )then
    if( offset_tail > offset_now )then
      call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                            f, little, ios, opt)
    endif

    if( cmn%imax_int1 > 0_8 )then  ! modified
      cmn%imax_int1 = 0_8
      deallocate(cmn%arr_int1)
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for2d_int1
!===============================================================
!
!===============================================================
subroutine write_row_for2d_int2(&
    dat, fill, byte, i1max, i2max, i2, &
    fsize, offset_set, offset_cur, &
    offset_tail, offset_now, offset_next, &
    f, dtype, little, ios, opt)
  implicit none
  integer(2)  , intent(in), target :: dat(:)  ! modified
  integer(2)  , intent(in)         :: fill  ! modified
  integer(8)  , intent(in)         :: byte
  integer(8)  , pointer            :: i1max
  integer(8)                       :: i2max, &
                                      i2
  integer(8)  , intent(in)    :: fsize
  integer(8)  , pointer       :: offset_set, &
                                 offset_cur
  integer(8)  , intent(in)    :: offset_tail
  integer(8)  , intent(inout) :: offset_now, &
                                 offset_next
  character(*), pointer     :: f
  character(*), intent(in)  :: dtype
  integer(4)  , pointer     :: little
  integer     , pointer     :: ios
  character(*), intent(in)  :: opt

  integer(2), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for2d_int2', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_next > fsize .and. offset_next > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_next, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    if( i2 == 1_8 )then
      call c_fseek_set(offset_set, ios)
      if( ios /= 0 )then
        call raise_error_fseek('set', f, offset_set, opt)
        call echo(CODE%RET)
        return
      endif
    else
      call c_fseek_cur(offset_cur, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', f, offset_cur, opt)
        call echo(CODE%RET)
        return
      endif
    endif
  endif

  offset_now = offset_next
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dat_ptr => dat

  call c_fwrite_int2(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  offset_now = offset_now + i1max * byte
  offset_next = offset_now + offset_cur
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == i2max )then
    if( offset_tail > offset_now )then
      call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                            f, little, ios, opt)
    endif

    if( cmn%imax_int2 > 0_8 )then  ! modified
      cmn%imax_int2 = 0_8
      deallocate(cmn%arr_int2)
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for2d_int2
!===============================================================
!
!===============================================================
subroutine write_row_for2d_int4(&
    dat, fill, byte, i1max, i2max, i2, &
    fsize, offset_set, offset_cur, &
    offset_tail, offset_now, offset_next, &
    f, dtype, little, ios, opt)
  implicit none
  integer(4)  , intent(in), target :: dat(:)  ! modified
  integer(4)  , intent(in)         :: fill  ! modified
  integer(8)  , intent(in)         :: byte
  integer(8)  , pointer            :: i1max
  integer(8)                       :: i2max, &
                                      i2
  integer(8)  , intent(in)    :: fsize
  integer(8)  , pointer       :: offset_set, &
                                 offset_cur
  integer(8)  , intent(in)    :: offset_tail
  integer(8)  , intent(inout) :: offset_now, &
                                 offset_next
  character(*), pointer     :: f
  character(*), intent(in)  :: dtype
  integer(4)  , pointer     :: little
  integer     , pointer     :: ios
  character(*), intent(in)  :: opt

  integer(4), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for2d_int4', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_next > fsize .and. offset_next > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_next, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    if( i2 == 1_8 )then
      call c_fseek_set(offset_set, ios)
      if( ios /= 0 )then
        call raise_error_fseek('set', f, offset_set, opt)
        call echo(CODE%RET)
        return
      endif
    else
      call c_fseek_cur(offset_cur, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', f, offset_cur, opt)
        call echo(CODE%RET)
        return
      endif
    endif
  endif

  offset_now = offset_next
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dat_ptr => dat

  call c_fwrite_int4(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  offset_now = offset_now + i1max * byte
  offset_next = offset_now + offset_cur
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == i2max )then
    if( offset_tail > offset_now )then
      call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                            f, little, ios, opt)
    endif

    if( cmn%imax_int4 > 0_8 )then  ! modified
      cmn%imax_int4 = 0_8
      deallocate(cmn%arr_int4)
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for2d_int4
!===============================================================
!
!===============================================================
subroutine write_row_for2d_int8(&
    dat, fill, byte, i1max, i2max, i2, &
    fsize, offset_set, offset_cur, &
    offset_tail, offset_now, offset_next, &
    f, dtype, little, ios, opt)
  implicit none
  integer(8)  , intent(in), target :: dat(:)  ! modified
  integer(8)  , intent(in)         :: fill  ! modified
  integer(8)  , intent(in)         :: byte
  integer(8)  , pointer            :: i1max
  integer(8)                       :: i2max, &
                                      i2
  integer(8)  , intent(in)    :: fsize
  integer(8)  , pointer       :: offset_set, &
                                 offset_cur
  integer(8)  , intent(in)    :: offset_tail
  integer(8)  , intent(inout) :: offset_now, &
                                 offset_next
  character(*), pointer     :: f
  character(*), intent(in)  :: dtype
  integer(4)  , pointer     :: little
  integer     , pointer     :: ios
  character(*), intent(in)  :: opt

  integer(8), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for2d_int8', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_next > fsize .and. offset_next > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_next, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    if( i2 == 1_8 )then
      call c_fseek_set(offset_set, ios)
      if( ios /= 0 )then
        call raise_error_fseek('set', f, offset_set, opt)
        call echo(CODE%RET)
        return
      endif
    else
      call c_fseek_cur(offset_cur, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', f, offset_cur, opt)
        call echo(CODE%RET)
        return
      endif
    endif
  endif

  offset_now = offset_next
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dat_ptr => dat

  call c_fwrite_int8(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  offset_now = offset_now + i1max * byte
  offset_next = offset_now + offset_cur
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == i2max )then
    if( offset_tail > offset_now )then
      call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                            f, little, ios, opt)
    endif

    if( cmn%imax_int8 > 0_8 )then  ! modified
      cmn%imax_int8 = 0_8
      deallocate(cmn%arr_int8)
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for2d_int8
!===============================================================
!
!===============================================================
subroutine write_row_for2d_real(&
    dat, fill, byte, i1max, i2max, i2, &
    fsize, offset_set, offset_cur, &
    offset_tail, offset_now, offset_next, &
    f, dtype, little, ios, opt)
  implicit none
  real(4)     , intent(in), target :: dat(:)  ! modified
  real(4)     , intent(in)         :: fill  ! modified
  integer(8)  , intent(in)         :: byte
  integer(8)  , pointer            :: i1max
  integer(8)                       :: i2max, &
                                      i2
  integer(8)  , intent(in)    :: fsize
  integer(8)  , pointer       :: offset_set, &
                                 offset_cur
  integer(8)  , intent(in)    :: offset_tail
  integer(8)  , intent(inout) :: offset_now, &
                                 offset_next
  character(*), pointer     :: f
  character(*), intent(in)  :: dtype
  integer(4)  , pointer     :: little
  integer     , pointer     :: ios
  character(*), intent(in)  :: opt

  real(4), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for2d_real', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_next > fsize .and. offset_next > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_next, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    if( i2 == 1_8 )then
      call c_fseek_set(offset_set, ios)
      if( ios /= 0 )then
        call raise_error_fseek('set', f, offset_set, opt)
        call echo(CODE%RET)
        return
      endif
    else
      call c_fseek_cur(offset_cur, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', f, offset_cur, opt)
        call echo(CODE%RET)
        return
      endif
    endif
  endif

  offset_now = offset_next
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dat_ptr => dat

  call c_fwrite_real(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  offset_now = offset_now + i1max * byte
  offset_next = offset_now + offset_cur
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == i2max )then
    if( offset_tail > offset_now )then
      call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                            f, little, ios, opt)
    endif

    if( cmn%imax_real > 0_8 )then  ! modified
      cmn%imax_real = 0_8
      deallocate(cmn%arr_real)
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for2d_real
!===============================================================
!
!===============================================================
subroutine write_row_for2d_dble(&
    dat, fill, byte, i1max, i2max, i2, &
    fsize, offset_set, offset_cur, &
    offset_tail, offset_now, offset_next, &
    f, dtype, little, ios, opt)
  implicit none
  real(8)     , intent(in), target :: dat(:)  ! modified
  real(8)     , intent(in)         :: fill  ! modified
  integer(8)  , intent(in)         :: byte
  integer(8)  , pointer            :: i1max
  integer(8)                       :: i2max, &
                                      i2
  integer(8)  , intent(in)    :: fsize
  integer(8)  , pointer       :: offset_set, &
                                 offset_cur
  integer(8)  , intent(in)    :: offset_tail
  integer(8)  , intent(inout) :: offset_now, &
                                 offset_next
  character(*), pointer     :: f
  character(*), intent(in)  :: dtype
  integer(4)  , pointer     :: little
  integer     , pointer     :: ios
  character(*), intent(in)  :: opt

  real(8), pointer :: dat_ptr(:)  ! modified

  call echo(CODE%BGN, 'write_row__MP__write_row_for2d_dble', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( offset_next > fsize .and. offset_next > offset_now )then
    call fill_empty_space(fsize, offset_now, offset_next, fill, &
                          f, little, ios, opt)
    if( ios /= 0 )then
      call echo(CODE%RET)
      return
    endif
  else
    if( i2 == 1_8 )then
      call c_fseek_set(offset_set, ios)
      if( ios /= 0 )then
        call raise_error_fseek('set', f, offset_set, opt)
        call echo(CODE%RET)
        return
      endif
    else
      call c_fseek_cur(offset_cur, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', f, offset_cur, opt)
        call echo(CODE%RET)
        return
      endif
    endif
  endif

  offset_now = offset_next
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  dat_ptr => dat

  call c_fwrite_dble(dat_ptr, i1max, little, ios)  ! modified

  if( ios /= 0 )then
    call raise_error_fwrite(dtype, f, i1max, little, opt)
    call echo(CODE%RET)
    return
  endif
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  offset_now = offset_now + i1max * byte
  offset_next = offset_now + offset_cur
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( i2 == i2max )then
    if( offset_tail > offset_now )then
      call fill_empty_space(fsize, offset_now, offset_tail, fill, &
                            f, little, ios, opt)
    endif

    if( cmn%imax_dble > 0_8 )then  ! modified
      cmn%imax_dble = 0_8
      deallocate(cmn%arr_dble)
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine write_row_for2d_dble
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine fill_empty_space_int1(fsize, offset_now, offset_next, fill, path, little, ios, opt)
  implicit none
  integer(8)  , intent(in) :: fsize
  integer(8)  , intent(in) :: offset_now
  integer(8)  , intent(in) :: offset_next
  integer(1)  , intent(in) :: fill  ! modified
  character(*), intent(in) :: path
  integer     , pointer    :: little
  integer     , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: imax

  integer :: d

  call echo(CODE%BGN, 'fill_empty__MP__fill_empty_space_int1', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( debug )then
    d = dgt(max(offset_now, offset_next, fsize))
    call edbg('offset now  : '//str(offset_now,d)//&
            '\n       next : '//str(offset_next,d)//&
            '\n       fsize: '//str(fsize,d))
  endif

  if( offset_next > fsize )then
    if( fsize > offset_now )then
      call c_fseek_cur(fsize-offset_now, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', path, fsize-offset_now, opt)
        call echo(CODE%RET)
        return
      endif
    endif

    imax = (offset_next - max(offset_now,fsize)) / 1_8  ! modified

    if( imax == 0_8 )then
      call eerr(str(msg_invalid_value())//&
              '\n  imax: '//str(imax)//&
              '\n  offset_now : '//str(offset_now,d)//&
              '\n  offset_next: '//str(offset_next,d)//&
              '\n  fsize      : '//str(fsize,d))
    endif

    ! modified
    if( imax /= cmn%imax_int1 )then
      if( cmn%imax_int1 > 0_8 ) deallocate(cmn%arr_int1)
      allocate(cmn%arr_int1(imax))
      cmn%imax_int1 = imax
    endif

    cmn%arr_int1(:) = fill

    call c_fwrite_int1(cmn%arr_int1, imax, little, ios)
    ! modified/

    if( ios /= 0 )then
      call raise_error_fwrite(DTYPE_INT1, path, size(cmn%arr_int1,kind=8), little, opt)  ! modified
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine fill_empty_space_int1
!===============================================================
!
!===============================================================
subroutine fill_empty_space_int2(fsize, offset_now, offset_next, fill, path, little, ios, opt)
  implicit none
  integer(8)  , intent(in) :: fsize
  integer(8)  , intent(in) :: offset_now
  integer(8)  , intent(in) :: offset_next
  integer(2)  , intent(in) :: fill  ! modified
  character(*), intent(in) :: path
  integer     , pointer    :: little
  integer     , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: imax

  integer :: d

  call echo(CODE%BGN, 'fill_empty__MP__fill_empty_space_int2', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( debug )then
    d = dgt(max(offset_now, offset_next, fsize))
    call edbg('offset now  : '//str(offset_now,d)//&
            '\n       next : '//str(offset_next,d)//&
            '\n       fsize: '//str(fsize,d))
  endif

  if( offset_next > fsize )then
    if( fsize > offset_now )then
      call c_fseek_cur(fsize-offset_now, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', path, fsize-offset_now, opt)
        call echo(CODE%RET)
        return
      endif
    endif

    imax = (offset_next - max(offset_now,fsize)) / 2_8  ! modified

    if( imax == 0_8 )then
      call eerr(str(msg_invalid_value())//&
              '\n  imax: '//str(imax)//&
              '\n  offset_now : '//str(offset_now,d)//&
              '\n  offset_next: '//str(offset_next,d)//&
              '\n  fsize      : '//str(fsize,d))
    endif

    ! modified
    if( imax /= cmn%imax_int2 )then
      if( cmn%imax_int2 > 0_8 ) deallocate(cmn%arr_int2)
      allocate(cmn%arr_int2(imax))
      cmn%imax_int2 = imax
    endif

    cmn%arr_int2(:) = fill

    call c_fwrite_int2(cmn%arr_int2, imax, little, ios)
    ! modified/

    if( ios /= 0 )then
      call raise_error_fwrite(DTYPE_INT2, path, size(cmn%arr_int2,kind=8), little, opt)  ! modified
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine fill_empty_space_int2
!===============================================================
!
!===============================================================
subroutine fill_empty_space_int4(fsize, offset_now, offset_next, fill, path, little, ios, opt)
  implicit none
  integer(8)  , intent(in) :: fsize
  integer(8)  , intent(in) :: offset_now
  integer(8)  , intent(in) :: offset_next
  integer(4)  , intent(in) :: fill  ! modified
  character(*), intent(in) :: path
  integer     , pointer    :: little
  integer     , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: imax

  integer :: d

  call echo(CODE%BGN, 'fill_empty__MP__fill_empty_space_int4', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( debug )then
    d = dgt(max(offset_now, offset_next, fsize))
    call edbg('offset now  : '//str(offset_now,d)//&
            '\n       next : '//str(offset_next,d)//&
            '\n       fsize: '//str(fsize,d))
  endif

  if( offset_next > fsize )then
    if( fsize > offset_now )then
      call c_fseek_cur(fsize-offset_now, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', path, fsize-offset_now, opt)
        call echo(CODE%RET)
        return
      endif
    endif

    imax = (offset_next - max(offset_now,fsize)) / 4_8  ! modified

    if( imax == 0_8 )then
      call eerr(str(msg_invalid_value())//&
              '\n  imax: '//str(imax)//&
              '\n  offset_now : '//str(offset_now,d)//&
              '\n  offset_next: '//str(offset_next,d)//&
              '\n  fsize      : '//str(fsize,d))
    endif

    ! modified
    if( imax /= cmn%imax_int4 )then
      if( cmn%imax_int4 > 0_8 ) deallocate(cmn%arr_int4)
      allocate(cmn%arr_int4(imax))
      cmn%imax_int4 = imax
    endif

    cmn%arr_int4(:) = fill

    call c_fwrite_int4(cmn%arr_int4, imax, little, ios)
    ! modified/

    if( ios /= 0 )then
      call raise_error_fwrite(DTYPE_INT4, path, size(cmn%arr_int4,kind=8), little, opt)  ! modified
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine fill_empty_space_int4
!===============================================================
!
!===============================================================
subroutine fill_empty_space_int8(fsize, offset_now, offset_next, fill, path, little, ios, opt)
  implicit none
  integer(8)  , intent(in) :: fsize
  integer(8)  , intent(in) :: offset_now
  integer(8)  , intent(in) :: offset_next
  integer(8)  , intent(in) :: fill  ! modified
  character(*), intent(in) :: path
  integer     , pointer    :: little
  integer     , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: imax

  integer :: d

  call echo(CODE%BGN, 'fill_empty__MP__fill_empty_space_int8', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( debug )then
    d = dgt(max(offset_now, offset_next, fsize))
    call edbg('offset now  : '//str(offset_now,d)//&
            '\n       next : '//str(offset_next,d)//&
            '\n       fsize: '//str(fsize,d))
  endif

  if( offset_next > fsize )then
    if( fsize > offset_now )then
      call c_fseek_cur(fsize-offset_now, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', path, fsize-offset_now, opt)
        call echo(CODE%RET)
        return
      endif
    endif

    imax = (offset_next - max(offset_now,fsize)) / 8_8  ! modified

    if( imax == 0_8 )then
      call eerr(str(msg_invalid_value())//&
              '\n  imax: '//str(imax)//&
              '\n  offset_now : '//str(offset_now,d)//&
              '\n  offset_next: '//str(offset_next,d)//&
              '\n  fsize      : '//str(fsize,d))
    endif

    ! modified
    if( imax /= cmn%imax_int8 )then
      if( cmn%imax_int8 > 0_8 ) deallocate(cmn%arr_int8)
      allocate(cmn%arr_int8(imax))
      cmn%imax_int8 = imax
    endif

    cmn%arr_int8(:) = fill

    call c_fwrite_int8(cmn%arr_int8, imax, little, ios)
    ! modified/

    if( ios /= 0 )then
      call raise_error_fwrite(DTYPE_INT8, path, size(cmn%arr_int8,kind=8), little, opt)  ! modified
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine fill_empty_space_int8
!===============================================================
!
!===============================================================
subroutine fill_empty_space_real(fsize, offset_now, offset_next, fill, path, little, ios, opt)
  implicit none
  integer(8)  , intent(in) :: fsize
  integer(8)  , intent(in) :: offset_now
  integer(8)  , intent(in) :: offset_next
  real(4)     , intent(in) :: fill  ! modified
  character(*), intent(in) :: path
  integer     , pointer    :: little
  integer     , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: imax
  integer :: d

  call echo(CODE%BGN, 'fill_empty__MP__fill_empty_space_real', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( debug )then
    d = dgt(max(offset_now, offset_next, fsize))
    call edbg('offset now  : '//str(offset_now,d)//&
            '\n       next : '//str(offset_next,d)//&
            '\n       fsize: '//str(fsize,d))
  endif

  if( offset_next > fsize )then
    if( fsize > offset_now )then
      call c_fseek_cur(fsize-offset_now, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', path, fsize-offset_now, opt)
        call echo(CODE%RET)
        return
      endif
    endif

    imax = (offset_next - max(offset_now,fsize)) / 4_8  ! modified

    if( imax == 0_8 )then
      call eerr(str(msg_invalid_value())//&
              '\n  imax: '//str(imax)//&
              '\n  offset_now : '//str(offset_now,d)//&
              '\n  offset_next: '//str(offset_next,d)//&
              '\n  fsize      : '//str(fsize,d))
    endif

    ! modified
    if( imax /= cmn%imax_real )then
      if( cmn%imax_real > 0_8 ) deallocate(cmn%arr_real)
      allocate(cmn%arr_real(imax))
      cmn%imax_real = imax
    endif

    cmn%arr_real(:) = fill

    call c_fwrite_real(cmn%arr_real, imax, little, ios)
    ! modified/

    if( ios /= 0 )then
      call raise_error_fwrite(DTYPE_REAL, path, size(cmn%arr_real,kind=8), little, opt)  ! modified
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine fill_empty_space_real
!===============================================================
!
!===============================================================
subroutine fill_empty_space_dble(fsize, offset_now, offset_next, fill, path, little, ios, opt)
  implicit none
  integer(8)  , intent(in) :: fsize
  integer(8)  , intent(in) :: offset_now
  integer(8)  , intent(in) :: offset_next
  real(8)     , intent(in) :: fill  ! modified
  character(*), intent(in) :: path
  integer     , pointer    :: little
  integer     , pointer    :: ios
  character(*), intent(in) :: opt

  integer(8) :: imax
  integer :: d

  call echo(CODE%BGN, 'fill_empty__MP__fill_empty_space_dble', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  if( debug )then
    d = dgt(max(offset_now, offset_next, fsize))
    call edbg('offset now  : '//str(offset_now,d)//&
            '\n       next : '//str(offset_next,d)//&
            '\n       fsize: '//str(fsize,d))
  endif

  if( offset_next > fsize )then
    if( fsize > offset_now )then
      call c_fseek_cur(fsize-offset_now, ios)
      if( ios /= 0 )then
        call raise_error_fseek('cur', path, fsize-offset_now, opt)
        call echo(CODE%RET)
        return
      endif
    endif

    imax = (offset_next - max(offset_now,fsize)) / 8_8  ! modified

    if( imax == 0_8 )then
      call eerr(str(msg_invalid_value())//&
              '\n  imax: '//str(imax)//&
              '\n  offset_now : '//str(offset_now,d)//&
              '\n  offset_next: '//str(offset_next,d)//&
              '\n  fsize      : '//str(fsize,d))
    endif

    ! modified
    if( imax /= cmn%imax_dble )then
      if( cmn%imax_dble > 0_8 ) deallocate(cmn%arr_dble)
      allocate(cmn%arr_dble(imax))
      cmn%imax_dble = imax
    endif

    cmn%arr_dble(:) = fill

    call c_fwrite_dble(cmn%arr_dble, imax, little, ios)
    ! modified/

    if( ios /= 0 )then
      call raise_error_fwrite(DTYPE_DBLE, path, size(cmn%arr_dble,kind=8), little, opt)  ! modified
      call echo(CODE%RET)
      return
    endif
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine fill_empty_space_dble
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine check_array_size(arr, sz, nam)
  implicit none
  integer(8)  , intent(in) :: arr(:)
  integer     , intent(in) :: sz
  character(*), intent(in) :: nam

  call echo(CODE%BGN, 'check_array_size', '-p')
  !-------------------------------------------------------------
  if( size(arr) /= sz )then
    call eerr('Size of $'//str(nam)//' must be '//str(sz)//'.')
  endif
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine check_array_size
!===============================================================
!
!===============================================================
subroutine check_values(rec, shp, sz, lb, ub, ios, opt)
  implicit none
  integer     , intent(in)  :: rec
  integer     , intent(in)  :: shp(:)
  integer(8)  , intent(in)  :: sz(:)
  integer(8)  , intent(in)  :: lb(:)
  integer(8)  , intent(in)  :: ub(:)
  integer     , intent(out) :: ios
  character(*), intent(in)  :: opt

  integer :: imax, i

  call echo(CODE%BGN, 'check_values', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  ios = 0

  if( rec < 1 )then
    if( rec == REC_UNDEF )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  rec == REC_UNDEF ('//str(REC_UNDEF)//')', opt)
    else
      call eerr(str(msg_invalid_value())//&
              '\n  rec: '//str(rec), opt)
    endif
    ios = 1
  endif

  imax = size(shp)
  if( size(sz) /= imax .or. size(lb) /= imax .or. size(ub) /= imax )then
    call eerr(str(msg_unexpected_condition())//&
           '\n  Size of arrays mismatch.'//&
           '\n  size(shp): '//str(size(shp))//&
           '\n  size(sz) : '//str(size(sz))//&
           '\n  size(lb) : '//str(size(lb))//&
           '\n  size(ub) : '//str(size(ub)), opt)
    ios = 1
  endif

  do i = 1, imax
    if( lb(i) < 1 )then
      call eerr(str(msg_invalid_value())//&
              '\n  lb('//str(i)//'): '//str(lb(i))//&
              '\nBoundaries must be positive.', opt)
      ios = 1
    endif

    if( ub(i) > sz(i) )then
      call eerr(str(msg_unexpected_condition())//&
              '\n  ub('//str(i)//') > sz('//str(i)//')'//&
              '\n  ub('//str(i)//') : '//str(ub(i))//&
              '\n  sz('//str(i)//') : '//str(sz(i))//&
              '\n  shp('//str(i)//'): '//str(shp(i)), opt)
      ios = 1
    endif
  enddo
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine check_values
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine alloc_ptr_1d(&
    dtype, imax, &
    arr_int1, arr_int2, arr_int4, arr_int8, arr_real, arr_dble)
  implicit none
  character(*), intent(in) :: dtype
  integer(8)  , intent(in) :: imax
  integer(1)  , pointer    :: arr_int1(:)
  integer(2)  , pointer    :: arr_int2(:)
  integer(4)  , pointer    :: arr_int4(:)
  integer(8)  , pointer    :: arr_int8(:)
  real(4)     , pointer    :: arr_real(:)
  real(8)     , pointer    :: arr_dble(:)

  call echo(CODE%BGN, 'alloc_ptr_1d', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  selectcase( dtype )
  case( DTYPE_INT1 )
    allocate(arr_int1(imax))
  case( DTYPE_INT2 )
    allocate(arr_int2(imax))
  case( DTYPE_INT4 )
    allocate(arr_int4(imax))
  case( DTYPE_INT8 )
    allocate(arr_int8(imax))
  case( DTYPE_REAL )
    allocate(arr_real(imax))
  case( DTYPE_DBLE )
    allocate(arr_dble(imax))
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype))
  endselect
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine alloc_ptr_1d
!===============================================================
!
!===============================================================
subroutine dealloc_ptr_1d(&
    dtype, &
    arr_int1, arr_int2, arr_int4, arr_int8, arr_real, arr_dble)
  implicit none
  character(*), intent(in) :: dtype
  integer(1)  , pointer    :: arr_int1(:)
  integer(2)  , pointer    :: arr_int2(:)
  integer(4)  , pointer    :: arr_int4(:)
  integer(8)  , pointer    :: arr_int8(:)
  real(4)     , pointer    :: arr_real(:)
  real(8)     , pointer    :: arr_dble(:)

  call echo(CODE%BGN, 'dealloc_ptr_1d', '-p')
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  selectcase( dtype )
  case( DTYPE_INT1 )
    deallocate(arr_int1)
  case( DTYPE_INT2 )
    deallocate(arr_int2)
  case( DTYPE_INT4 )
    deallocate(arr_int4)
  case( DTYPE_INT8 )
    deallocate(arr_int8)
  case( DTYPE_REAL )
    deallocate(arr_real)
  case( DTYPE_DBLE )
    deallocate(arr_dble)
  case default
    call eerr(str(msg_invalid_value())//&
            '\n  dtype: '//str(dtype))
  endselect
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine dealloc_ptr_1d
!===============================================================
!
!===============================================================
!
!
!
!
!
!===============================================================
!
!===============================================================
subroutine raise_error_fopen(path, opt)
  implicit none
  character(*), intent(in) :: path
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'raise_error_fopen', '-p')
  !-------------------------------------------------------------
  call eerr('An error occured in function "c_fopen".'//&
           '\n  path: '//str(path), opt)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine raise_error_fopen
!===============================================================
!
!===============================================================
subroutine raise_error_fclose(path, opt)
  implicit none
  character(*), intent(in) :: path
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'raise_error_fopen', '-p')
  !-------------------------------------------------------------
  call eerr('An error occured in function "c_fclose".'//&
          '\n  path: '//str(path), opt)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine raise_error_fclose
!===============================================================
!
!===============================================================
subroutine raise_error_fseek(nam, path, offset, opt)
  implicit none
  character(*), intent(in) :: nam
  character(*), intent(in) :: path
  integer(8)  , intent(in) :: offset
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'raise_error_fseek', '-p')
  !-------------------------------------------------------------
  call eerr('An error occured in function "c_fseek_'//str(nam)//'".'//&
           '\n  path  : '//str(path)//&
           '\n  offset: '//str(offset), opt)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine raise_error_fseek
!===============================================================
!
!===============================================================
subroutine raise_error_fread(nam, path, nij, little, opt)
  implicit none
  character(*), intent(in) :: nam
  character(*), intent(in) :: path
  integer(8)  , intent(in) :: nij
  integer     , intent(in) :: little
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'raise_error_fread', '-p')
  !-------------------------------------------------------------
  call eerr('An error occured in "c_fread_'//str(nam)//'".'//&
          '\n  path  : '//str(path)//&
          '\n  nij   : '//str(nij)//&
          '\n  little: '//str(little), opt)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine raise_error_fread
!===============================================================
!
!===============================================================
subroutine raise_error_fwrite(nam, path, nij, little, opt)
  implicit none
  character(*), intent(in) :: nam
  character(*), intent(in) :: path
  integer(8)  , intent(in) :: nij
  integer     , intent(in) :: little
  character(*), intent(in) :: opt

  call echo(CODE%BGN, 'raise_error_fwrite', '-p')
  !-------------------------------------------------------------
  call eerr('An error occured in "c_fwrite_'//str(nam)//'".'//&
          '\n  path  : '//str(path)//&
          '\n  nij   : '//str(nij)//&
          '\n  little: '//str(little), opt)
  !-------------------------------------------------------------
  call echo(CODE%RET)
end subroutine raise_error_fwrite
!===============================================================
!
!===============================================================
end module lib_io_binary
